from __future__ import absolute_import

import xodb
from xodb import Attribute
from xodb.backends.xapian import XapianSchema, XapianAttribute
import datetime

class Ding(object):
    
    def __init__(self, bar):
        self.bar = bar
        self.blig = 41
        self.bloog = 42

        self.language = 'en'
        self.name = 'ding'
        self.direction = 'north'
        self.title = 'The young and the database.'
        self.instructions = 'Under the walker'
        self.zig = 3
#        self.creation_date = datetime.datetime(1999, 12, 31)


class DingSchema(XapianSchema):

    bar  = Attribute(type=Bar)
    zig  = Attribute(required=False)
    blig = Attribute(volatile=True)

    name          = XapianAttribute(text=True)
    direction     = XapianAttribute(prefixed=True)
    title         = XapianAttribute(text=True, language='fr')
    instructions  = XapianAttribute(prefixed=True, text=True, language='es')
#    creation_date = Attribute(Value())





from __future__ import absolute_import

import xapian
import string

from cStringIO import StringIO
from cPickle import Pickler, Unpickler
from xapian import QueryParser, Query, UnimplementedError, DocNotFoundError

from . import snowball
from .spies import TermCountMatchSpy


RETRY_LIMIT = 3

try:
    # try bz2, fallback to zlib
    from bz2 import compress, decompress

    def _decompress(data):
        if data.startswith('BZ'):
            try:
                data = decompress(data)
            except IOError:
                pass
        return data

except ImportError:
    from zlib import compress, decompress, error as compress_error

    def _decompress(data):
        if data.startswith('x'):
            try:
                data = decompress(data)
            except compress_error:
                pass
        return data


from . schema import Schema


default_parser_flags = (QueryParser.FLAG_PHRASE |
                        QueryParser.FLAG_BOOLEAN |
                        QueryParser.FLAG_LOVEHATE |
                        QueryParser.FLAG_SPELLING_CORRECTION |
                        QueryParser.FLAG_BOOLEAN_ANY_CASE |
                        QueryParser.FLAG_WILDCARD)


class LangDecider(xapian.ExpandDecider):
    """
    A Xapian ExpandDecider that decide which terms to keep and which
    to discard when expanding a query using the "suggest" syntax.  As
    a place to start, we throw out:

      - Terms that don't begin with an uppercase letter or digit.
        This filters prefixed terms and stemmed forms.

      - Terms shorter than min_length chars, which are likely irrelevant

      - Stopwords for the given language.  Default is english, pass
        None for the lang argument if no stopping is desired.
    """

    min_length = 5
    nostart = unicode(string.uppercase + string.digits)

    def __init__(self, lang="en", filter=None, stems=None):
        super(LangDecider, self).__init__()
        if lang in snowball.stoppers:
            self.stopper = snowball.stoppers[lang]
            self.stemmer = xapian.Stem(lang)
        else:
            self.stopper = lambda(term): False
            self.stemmer = xapian.Stem("none")
        self.stems = tuple(self.stemmer(t) for t in stems) if stems else ()

    def __call__(self, term):
        term = term.decode("utf-8")
        if (term[0] in self.nostart or
            len(term) < self.min_length or
            self.stopper(term) or
            '_' in term or
            self.stemmer(term) in self.stems):
            return False
        return True


class PrefixDecider(xapian.ExpandDecider):

    __slots__ = ['prefix']

    def __init__(self, prefix):
        super(PrefixDecider, self).__init__()
        self.prefix = 'X' + prefix.upper()

    def __call__(self, term):
        return term.decode('utf-8').startswith(self.prefix)


class MultipleValueRangeProcessor(xapian.ValueRangeProcessor):
    """Value range processor for multiple prefixes.

    :param map: a dict of prefix to value number pairs.

    :param serializer: optional callable to serialize the arguments into
                       the same form as the corresponding values are stored.
                       Typically xapian.sortable_serialise for floats,
                       str.lower for strings.
    """

    def __init__(self, map, serializer=None):
        self.map = map
        self.serializer = serializer or (lambda x: x)
        xapian.ValueRangeProcessor.__init__(self)

    def __call__(self, begin, end):
        for prefix, value in self.map.items():
            if begin.startswith(prefix + ':'):
                return (value,
                        self.serializer(begin[len(prefix) + 1:]),
                        self.serializer(end))
        return (xapian.BAD_VALUENO, begin, end)


class Record(object):
    """Default dumb record type, names begining with _xodb_ are reserved ."""

    def __init__(self, document, otype, newargs, state):
        # this does not mangle the names
        state.update(__xodb_document=document,
                     __xodb_type=otype,
                     __xodb_newargs=newargs)
        self.__dict__ = state


class BrokenRecord(Record):
    """Type returned when pickled type cannot be found."""
    pass


def instance_factory(document, otype, newargs, state):
    inst = otype.__new__(otype, *newargs)
    inst.__dict__ = state
    return inst


class Database(object):
    """A xodb database.

    :param db_or_path: A path to file, or a xapian.Database object
    that backs this Database instance.

    :param writable: Open database in writable mode.

    :param overwrite: If writable is True, overwrite the existing
    database with a new one.

    :param spelling: If True, write spelling correction data to the
    database.

    :param record_factory: Callable factory to pass stored object
    state when results are instanciated.

    :param persistent_id: Callable to resolve persistent ids for
    pickled object state.

    :param persistent_load: Callable to resolve pickled object state
    for persistent id.
    """
    COMPRESS_OVER = 256 # set to 0 to disable compression

    relevance_prefix = "_XODB_RP_"
    boolean_prefix = "_XODB_BP_"
    value_prefix = "_XODB_VALUE_"
    value_sort_prefix = "_XODB_VALUESORT_"
    value_count_name = "_XODB_COUNT_"

    def __init__(self, db_or_path,
                 writable=True, overwrite=False,
                 spelling=True, record_factory=Record,
                 persistent_id=None, persistent_load=None):
        self._writable = writable
        self.backend = db_or_path
        self.spelling = spelling
        self.record_factory = record_factory
        self.persistent_id = persistent_id
        self.persistent_load = persistent_load

        if isinstance(db_or_path, basestring):
            if writable:
                if overwrite:
                    flags = xapian.DB_CREATE_OR_OVERWRITE
                else:
                    flags = xapian.DB_CREATE_OR_OPEN
                self.backend = xapian.WritableDatabase(db_or_path, flags)
            else:
                self.backend = xapian.Database(db_or_path)
        elif not isinstance(db_or_path, xapian.Database):
            raise TypeError(
                'First argument must be path or xapian.[Writable]Database')

        self.relevance_prefixes = {}
        self.boolean_prefixes = {}
        self.values = {}
        self.value_sorts = {}
        self.type_map = {}

        self.map(object, Schema)
        self.reopen()

    def map(self, otype, schema):
        """Map a type to a schema."""
        self.type_map[otype] = schema

    def unmap(self, otype):
        """Unmap a type."""
        del self.type_map[otype]

    def clear_map(self):
        self.type_map = {}

    def schema_for(self, otype):
        """Get the schema for a given type, or one of its
        superclasses."""
        for base in otype.__mro__:
            if base in self.type_map:
                return self.type_map[base]
        return Schema

    def serialize(self, typ, newargs, state):
        f = StringIO()
        pickler = Pickler(f, -1)
        pickler.dump((typ, newargs))
        if self.persistent_id is not None:
            pickler.persistent_id = self.persistent_id
        pickler.dump(state)
        data = f.getvalue()
        if self.COMPRESS_OVER and len(data) > self.COMPRESS_OVER:
            data = compress(data)
        return data

    def unserialize(self, document, record_factory=None):
        
        data = _decompress(document.get_data())
        file = StringIO(data)
        unpickler = Unpickler(file)
        try:
            otype, newargs = unpickler.load()
        except AttributeError:
            # the type may have gone away ...
            otype, newargs = BrokenRecord, ()
        if self.persistent_load is not None:
            unpickler.persistent_load = self.persistent_load
        state = unpickler.load()
        if record_factory:
            return record_factory(document, otype, newargs, state)
        else:
            return document, otype, newargs, state

    def _get_value_count(self):
        return int(self.backend.get_metadata(self.value_count_name))

    def _set_value_count(self, count):
        self.backend.set_metadata(self.value_count_name, str(count))

    value_count = property(_get_value_count, _set_value_count)

    def add_prefix(self, key, value):
        self.relevance_prefixes[key] = value
        self.backend.set_metadata(self.relevance_prefix + key, value)

    def add_boolean_prefix(self, key, value):
        self.boolean_prefixes[key] = value
        self.backend.set_metadata(self.boolean_prefix + key, value)

    def add_value(self, name, sort=None):
        if name in self.values:
            return self.values[name]
        value_count = self.value_count + 1
        self.value_count = value_count
        self.values[name] = value_count
        self.backend.set_metadata(self.value_prefix + name, str(value_count))
        if sort:
            self.value_sorts[name] = sort
            self.backend.set_metadata(self.value_sort_prefix + name, sort)
        return value_count

    def __len__(self):
        """ Return the number of documents in this database. """
        self.backend.reopen()
        return self.backend.get_doccount()

    def allterms(self, prefix=""):
        for t in self.backend.allterms(prefix):
            yield t.term

    def get(self, docid, default=None):
        """ Get a document with the given docid, or the default value if
        no such document exists.
        """
        try:
            return self[docid]
        except DocNotFoundError:
            return default

    def __getitem__(self, docid):
        document = self.backend.get_document(docid)
        return self.unserialize(document, self.record_factory)

    def __delitem__(self, docid):
        self.backend.delete_document(docid)

    def __setitem__(self, docid, document):
        doc = self.get(docid)
        if doc is None:
            self.backend.add_document(document)
        else:
            self.backend.replace_document(docid, document)

    def add(self, *objs):
        added = []
        for obj in objs:
            if hasattr(obj, '__xodb_schema__'):
                schema_type = obj.__xodb_schema__
            else:
                schema_type = self.schema_for(type(obj))
            schema = schema_type(self)
            added.append(self.backend.add_document(schema(obj)))
        return added

    def reopen(self):
        self.backend.reopen()
        try:
            keys = self.backend.metadata_keys()
        except UnimplementedError:
            # Inmemory backends don't expose this function, so just
            # return since they can't persist anything anyway!
            return

        self.relevance_prefixes.clear()
        self.boolean_prefixes.clear()
        self.values.clear()
        self.value_sorts.clear()

        for k in keys:
            if k.startswith(self.relevance_prefix):
                prefix = k[len(self.relevance_prefix):]
                self.relevance_prefixes[prefix] = self.backend.get_metadata(k)
            elif k.startswith(self.boolean_prefix):
                prefix = k[len(self.boolean_prefix):]
                self.boolean_prefixes[prefix] = self.backend.get_metadata(k)
            elif k.startswith(self.value_prefix):
                value = k[len(self.value_prefix):]
                self.values[value] = int(self.backend.get_metadata(k))
            elif k.startswith(self.value_sort_prefix):
                value = k[len(self.value_sort_prefix):]
                self.value_sorts[value] = self.backend.get_metadata(k)

        # hit the property to refresh this value
        try:
            count = self.value_count
        except ValueError:
            if self._writable:
                self.value_count = 0

    def begin(self):
        if self._writable:
            self.backend.reopen()
            self.backend.begin_transaction()

    def cancel(self):
        if self._writable:
            self.backend.cancel_transaction()

    def commit(self):
        if self._writable:
            self.backend.commit_transaction()

    def prepare_query_parser(self, lang=None, default_op=xapian.Query.OP_AND):
        qp = xapian.QueryParser()
        qp.set_database(self.backend)
        qp.set_default_op(default_op)

        if self.relevance_prefixes:
            for key, value in self.relevance_prefixes.items():
                qp.add_prefix(key, value)
        if self.boolean_prefixes:
            for key, value in self.boolean_prefixes.items():
                qp.add_boolean_prefix(key, value)
        if self.value_sorts:
            # First add numeric values ranges
            qp.add_valuerangeprocessor(MultipleValueRangeProcessor(
                dict(((k, self.values[k])
                      for k, v in self.value_sorts.items() if v == 'numeric')),
                lambda s: xapian.sortable_serialise(float(s)),
            ))
            # Then string and date
            qp.add_valuerangeprocessor(MultipleValueRangeProcessor(
                dict(((k, self.values[k])
                      for k, v in self.value_sorts.items() if v != 'numeric')),
            ))
        if lang in snowball.stoppers:
            qp.set_stemmer(xapian.Stem(lang))
            qp.set_stopper(snowball.stoppers[lang])
            qp.set_stemming_strategy(xapian.QueryParser.STEM_SOME)
        return qp

    def query(self, query, offset=0, limit=None, order=None,
              reverse=False, lang=None, check=0,
              match_decider=None, match_spy=None,
              echo=False, parser_flags=default_parser_flags,
              document=False,
              default_op=xapian.Query.OP_AND):
        """
        Query the database with the provided string or xapian Query
        object.  A string is passed into xapians QueryParser first to
        generate a Query object.
        """
        self.reopen()
        if query == "":
            query = xapian.Query("")
        else:
            qp = self.prepare_query_parser(lang, default_op)
            query = qp.parse_query(query, parser_flags)

        if echo:
            print str(query)
        enq = xapian.Enquire(self.backend)
        enq.set_query(query)

        mset = self._build_mset(enq, offset, limit, order, reverse,
                                check, match_decider, match_spy)
        return self._return_objects(mset, document=document)

    def estimate(self, query, limit=0, lang=None,
                 parser_flags=default_parser_flags):
        """Estimate the number of documents that will be yielded with the
        given query.

        Limit tells the estimator the minimum number of documents to
        consider.  A zero limit means check all documents in the db."""
        self.reopen()
        enq = xapian.Enquire(self.backend)

        if limit == 0:
            limit = self.backend.get_doccount()

        if query == "":
            query = xapian.Query("")
        else:
            qp = self.prepare_query_parser(lang)
            query = qp.parse_query(query, parser_flags)

        enq.set_query(query)
        return enq.get_mset(0, 0, limit).get_matches_estimated()

    def term_freq(self, term):
        """
        Return a count of the number of documents indexed for a given
        term.  Useful for testing.
        """
        self.backend.reopen()
        return self.backend.get_termfreq(term)

    def get_doccount(self):
        """
        Return the number of indexed documents, handy for tests and
        sanity check.
        """
        self.backend.reopen()
        return self.backend.get_doccount()

    def describe_query(self, query, lang=None,
              default_op=xapian.Query.OP_AND):
        """
        Describe the parsed query.
        """
        qp = self.prepare_query_parser(lang, default_op)
        q = qp.parse_query(query, default_parser_flags)
        return str(q)

    def spell(self, query, lang=None):
        """
        Suggest a query string with corrected spelling.
        """
        self.backend.reopen()
        qp = self.prepare_query_parser(lang)
        qp.parse_query(query, xapian.QueryParser.FLAG_SPELLING_CORRECTION)
        return qp.get_corrected_query_string().decode('utf8')

    def suggest(self, query, offset=0, limit=10, moffset=0, mlimit=10,
                lang=None, prefix=None, decider=None, score=False):
        """
        Suggest terms that would possibly yield more relevant results
        for the given query.
        """
        self.reopen()
        enq = xapian.Enquire(self.backend)

        qp = self.prepare_query_parser(lang)

        if mlimit is None:
            mlimit = self.backend.get_doccount()

        enq.set_query(qp.parse_query(query))

        mset = self._build_mset(enq, offset=moffset, limit=mlimit)

        rset = xapian.RSet()
        for m in mset:
            if hasattr(m, 'docid'): # 1.1
                docid = m.docid
            else:
                docid = m[xapian.MSET_DID] # 1.0
            rset.add_document(docid)

        if prefix is not None:
            decider = PrefixDecider(prefix)

        if decider is None:
            decider = LangDecider(lang)
        eset = enq.get_eset(limit, rset, decider)

        for item in eset.items:
            val = item[0].decode('utf8')
            if prefix and val.startswith('X' + prefix.upper()):
                val = "%s:%s" % (prefix, val[len(prefix) + 1:])
            if score:
                yield (val, item[1])
            else:
                yield val

    def similar(self, query=None, offset=0, limit=None, order=None,
                reverse=False, lang=None, check=None,
                match_decider=None, match_spy=None, terms=None,
                default_op=xapian.Query.OP_AND):
        """ Find documents in the database most relevant to the given terms.

        'query' - If *terms*=None, query is passed to xaql.suggest to
        get a list of terms. If *terms* is set, that term list is used
        directly and the *query* is parsed and ANDed with the elite term
        set query.

        'limit' - The number of records to return.

        'lang' - The language stemmer the parser should use.

        'match_decider' - A MatchDecider to apply to the query.

        'terms' - Optional term list from which to derive the
        elite set of terms to match on.
        """
        self.reopen()

        if not terms:
            assert query, 'Query or terms is requred'
            suggested_terms = self.suggest(
                query, lang=lang)
            terms = list(suggested_terms)
            query = None
        elif query:
            qp = self.prepare_query_parser(lang, default_op)
            query = qp.parse_query(query, default_parser_flags)

        enq = xapian.Enquire(self.backend)

        if limit is None:
            limit = self.backend.get_doccount()

        similar_query = Query(Query.OP_ELITE_SET, terms, limit)
        if query:
            similar_query = Query(Query.OP_AND, [similar_query, query])
        enq.set_query(similar_query)

        mset = self._build_mset(enq, offset, limit, order, reverse,
                                check, match_decider, match_spy)
        return self._return_objects(mset)

    def _build_mset(self, enq, offset=0, limit=None, order=None, reverse=False,
                    check=None, match_decider=None, match_spy=None):

        if order is not None:
            if isinstance(order, basestring):
                try:
                    order = self.values[order]
                except KeyError:
                    raise ValueError("There is no sort name %s" % order)
            enq.set_sort_by_value(order, reverse)

        if limit is None:
            limit = self.backend.get_doccount()

        if check is None:
            check = limit + 1

        tries = 0
        while True:
            try:
                mset = enq.get_mset(
                    offset, limit, check, None, match_decider, match_spy)
                break
            except xapian.DatabaseModifiedError:
                if tries > RETRY_LIMIT:
                    raise
                self.reopen()
                tries += 1
        return mset

    def _return_objects(self, mset, document=False):
        for record in mset:
            if hasattr(record, 'document'):  # 1.1
                doc = record.document
            else:
                doc = record[xapian.MSET_DOCUMENT] # 1.0
            if document:
                yield doc
            else:
                obj = self.unserialize(doc, record_factory=self.record_factory)
                yield obj

    def term_counter(self, prefixes):
        """Construct a term count spy with this instance's prefix dict."""
        prefix_map = self.relevance_prefixes.copy()
        prefix_map.update(self.boolean_prefixes)
        return TermCountMatchSpy(prefixes, prefix_map)



import calendar
import datetime
import operator
import math
from cPickle import loads

from xapian import MatchDecider


class TermCountMatchSpy(MatchDecider):
    """Pure python implementation of term counting match decider that
    should appear in the xapian bindings with the release of 1.1.

    'prefixes': iterable of prefix names strings

    'prefix_dictionary': mapping of prefix names to short prefixes
    normally passed in by Xaql when calling Xaql.term_counter
    """

    def __init__(self, prefixes, prefix_dictionary):
        MatchDecider.__init__(self)
        self.documents_seen = 0
        self.terms_seen = 0
        self.terms = dict((k, {}) for k in prefixes)
        self.prefix_dictionary = prefix_dictionary

    def __call__(self, doc):
        self.documents_seen += 1

        for prefix_name in self.terms:
            terms = self.terms[prefix_name]
            if prefix_name not in self.prefix_dictionary:
                continue
            prefix_code = self.prefix_dictionary[prefix_name]
            prefix_len = len(prefix_code)
            for term in doc:
                if prefix_code in term.term[0:prefix_len]:
                    self.terms_seen += 1
                    term = term.term[prefix_len:]
                    terms[term] = terms.get(term, 0) + 1
        return True

    def get_top_terms(self, prefix, maxterms=10):
        return dict(sorted(
            self.terms[prefix].items(),
            key=operator.itemgetter(1),
            reverse=True,
        )[0:maxterms])


class SchemaValueCountMatchSpy(MatchDecider):
    """Count values in xapian based on a schema.

    :param schema: The schema that defines which values to count, and
    whether they are atoms, sequences, or mappings.

    :param backend: The xapian backend used to resolve value name to
    number mapping.
    """

    def __init__(self, schema, database, facet_on=None):
        MatchDecider.__init__(self)
        self.documents_seen = 0
        self.values_seen = 0
        self.values = dict()
        self.schema = schema
        self.database = database
        self.facet_on = facet_on

        for name, attr in schema.__attributes__.items():
            if attr.value_field:
                if facet_on:
                    if attr.name not in facet_on:
                        continue
                self.values[name] = {}

    def _tally(self, valname, value):
        if value:
            vd = self.values[valname]
            vd[value] = vd.get(value, 0) + 1
            self.values_seen += 1

    def __call__(self, doc):
        self.documents_seen += 1

        for valname in self.values:
            attr = self.schema.__attributes__[valname]
            valno = self.database.values[valname]
            value = doc.get_value(valno)
            if not value:
                continue
            if not attr.sortable:
                value = loads(value)
            if value:
                if attr.sequence:
                    for v in value:
                        self._tally(valname, v)
                elif attr.mapping:
                    pass # not sure...
                else:
                    self._tally(valname, value)
        return True

    def top_values(self, valname, maxvalues=None):
        top_values = sorted(
            self.values[valname].items(),
            key=operator.itemgetter(1),
            reverse=True,
        )
        if maxvalues:
            top_values = top_values[0:maxvalues]
        return top_values

    def score_categorization(self, valname, desired_categories=None):
        """Return a scoring on this value with the desired number
        categories.  Lower is better.

        Added the values_seen concept here and in the ValueCountMatchSpy
        because it seems that the current implementation in c++ in the
        matchspy branch is severely broken for multivalued values.
        """
        score = 0.0
        categories = self.values[valname]
        desired_categories = desired_categories or len(categories)
        avg = float(self.values_seen) / desired_categories
        uncounted = self.values_seen

        for category, count in categories.items():
            uncounted -= count
            score += math.pow(count - avg, 2)
        assert uncounted >= 0, "Uncounted should never go negative"
        if uncounted:
            score += math.pow(uncounted - avg, 2)

        # normalize
        score /= math.pow(self.values_seen, 2)

        # bias towards exact number of categories requested
        score += 0.01 * math.pow(desired_categories - len(categories), 2)

        return score

    def build_numeric_ranges(self, valname, max_ranges, quantize=None):
        """
        Given a value name *valname*, the desired number of ranges
        and a quantization factor, returns a set of (low,high) range
        tuples to cover the range of values present.

        You may get back fewer than *max_ranges* if the quantization
        doesn't fit nicely.

        When no quantization factor is supplied a *max_ranges* ranges
        are return such that the value are evenly distributed accross them.
        (As in the C++ implementation in the match spies branch.)
        """

        if max_ranges < 1:
            return []

        # Make a list of all the values (with copies according to freq)
        values = [([float(v)] * f)
                  for v, f in self.values[valname].items()]
        # Collapse sub-lists and sort
        values = reduce(operator.add, sorted(values))

        count = len(values)
        bucket_size = int(count / max_ranges)
        if count % max_ranges and bucket_size > 1:
            bucket_size += 1


        quant_floor = lambda value, quant: value - (value % quant)
        quant_ceil = lambda value, quant: value if not (value % quant) \
            else value + (quant - value % quant)

        buckets = []
        if quantize:
            bottom = quant_floor(values[0], quantize)
            top = quant_ceil(values[-1], quantize)
            _range = top - bottom

            steps = int(math.floor(_range / float(quantize)))
            if max_ranges > steps:
                max_ranges = steps
            step_size = quant_floor(_range / max_ranges, quantize)

            base = bottom
            for n in range(max_ranges-1):
                buckets.append((base, base + step_size))
                base += step_size
            buckets.append((base, top))
        else:
            starts = map(lambda v: (v * bucket_size), range(0, max_ranges))
            ends = map(lambda v: (v + bucket_size), starts)
            ends[-1] = count
            buckets = [(values[s], values[e-1]) for s, e in zip(starts, ends)]

        return buckets

    def build_date_ranges(self, valname=None, future=False,
                          from_date=None, months=12):
        """Return a set of graduated date ranges forward or back from
        *from_date*.

        :param valname: Optional value slot name.  When provided, any
                        facets not matching values in the search
                        results will be filtered.  TODO:sw implement
                        the filtering.

        :param future: Defaults to False.  When true facets go forward
                       in time.  When false the go back in time.

        :param from_date: Defaults to today.
        """

        today = from_date or datetime.date.today()
        format = lambda d: d.strftime("%Y%m%d")
        last_day = lambda y, m: 29 if \
                   (calendar.isleap(y) and m == 2) else calendar.mdays[m]
        ranges = []
        if future:
            deltas = (('today', 0, 0), ('tomorrow', 1, 1), ('this week', 2, 7))
            months = xrange(months)
        else:
            deltas = (('today', 0, 0),
                      ('yesterday', -1, -1),
                      ('past week', -2, -7))
            months = xrange(0, -months, -1)
        for d in deltas:
            ranges.append((
                format(today + datetime.timedelta(d[1])),
                format(today + datetime.timedelta(d[2])),
                d[0],
            ))
        for n in months:
            mm = (today.month - 1 + n)
            month = (mm % 12) + 1
            year = mm / 12 + today.year
            ranges.append((
                format(datetime.date(year, month, 1)),
                format(datetime.date(year, month, last_day(year, month))),
                "%s %d" % (calendar.month_name[month], year),
            ))
        for range in ranges[:]:
            if valname:
                values = self.values[valname]
                for value in values:
                    if value >= range[0] and value <= range[1]:
                        break
                else:
                    ranges.remove(range)
        return ranges



from __future__ import absolute_import

class lazy_property(object):
    def __init__(self, fn):
        self.fn = fn
        self.__name__ = fn.func_name
        self.__doc__ = fn.__doc__

    def __get__(self, obj, cls):
        if obj is None:
            return None
        obj.__dict__[self.__name__] = result = self.fn(obj)
        return result



import cPickle as pickle
import datetime
from xodb.spies import *
import xodb
from xapian import Document
from xodb import Attribute, Schema

class AS(Schema):
    id = Attribute()
    bone = Attribute(value=True)
    wall = Attribute(value=True)
    seq = Attribute(value=True, sequence=True)


def test_value_spy():
    N = 15
    db = xodb.inmemory()
    ctr = SchemaValueCountMatchSpy(AS, db)

    valnos = dict((a.name, db.add_value(a.name)) for a in (AS.bone, AS.wall, AS.seq))

    words = ('foo', 'bar', 'baz', 'zoom')
    for i in range(N):
        doc = Document()
        doc.add_value(valnos['bone'], pickle.dumps('blong'))
        doc.add_value(valnos['wall'], pickle.dumps(words[i % len(words)]))
        ctr(doc)

    assert 'blong' in ctr.values['bone']
    assert ctr.values['bone']['blong'] == N
    assert ('blong', 15) in ctr.top_values('bone')

    for word in words:
        assert word in ctr.values['wall']

    assert ctr.values['wall']['foo'] == 4
    assert ctr.values['wall']['zoom'] == 3

    assert len(ctr.top_values('wall', 3)) == 3
    assert 'zoom' not in dict(ctr.top_values('wall', 3))

    assert ('zoom', 3) in ctr.top_values('wall')
    assert ('bar', 4) in ctr.top_values('wall')

def test_value_spy_multi():
    db = xodb.inmemory()
    ctr = SchemaValueCountMatchSpy(AS, db)

    valnos = dict((a.name, db.add_value(a.name)) for a in (AS.bone, AS.wall, AS.seq))

    words = ('foo', 'bar', 'baz', 'zoom')
    for i in range(len(words)+1):
        doc = Document()
        value = pickle.dumps(words[:i])
        doc.add_value(valnos['seq'], value)
        ctr(doc)

    for word in words:
        assert word in ctr.values['seq']

    assert ctr.values['seq']['foo'] == 4
    assert ctr.values['seq']['zoom'] == 1

    assert ctr.top_values('seq') == [('foo', 4), ('bar', 3), ('baz', 2), ('zoom', 1)]

def test_category_spy_exact():
    db = xodb.inmemory()
    ctr = SchemaValueCountMatchSpy(AS, db)

    valnos = dict((a.name, db.add_value(a.name)) for a in (AS.bone, AS.wall, AS.seq))

    words = ('foo', 'bar', 'baz', 'zoom')
    for i in range(len(words)):
        doc = Document()
        value = pickle.dumps(words[i])
        doc.add_value(valnos['bone'], value)
        ctr(doc)

    assert ctr.score_categorization('bone') == 0.0
    assert ctr.score_categorization('bone',4) == 0.0
    assert ctr.score_categorization('bone',3) > 0.0

def test_category_spy_approx():
    db = xodb.inmemory()
    ctr = SchemaValueCountMatchSpy(AS, db)

    valnos = dict((a.name, db.add_value(a.name)) for a in (AS.bone, AS.wall, AS.seq))

    words = ('foo', 'bar', 'baz', 'zoom')
    doc = Document()
    value = pickle.dumps(words)
    doc.add_value(valnos['seq'], value)
    ctr(doc)

    assert ctr.score_categorization('seq',4) < ctr.score_categorization('seq',5)
    assert ctr.score_categorization('seq',4) < ctr.score_categorization('seq',3)

def test_category_numeric_ranges():
    db = xodb.inmemory()
    ctr = SchemaValueCountMatchSpy(AS, db)

    valnos = dict((a.name, db.add_value(a.name)) for a in (AS.bone, AS.wall, AS.seq))

    values = (1.0, 2.0, 3.0, 3.3, 3.6, 3.9, 4.0, 5.0)
    for i in range(len(values)):
        doc = Document()
        value = pickle.dumps(unicode(values[i]))
        doc.add_value(valnos['bone'], value)
        ctr(doc)

    assert ctr.build_numeric_ranges('bone', 0) == []

    for N in range(1,len(values)+1):
        buckets = ctr.build_numeric_ranges('bone', N)
        assert len(buckets) == N
        assert buckets[0][0] == values[0]
        assert buckets[-1][1] == values[-1]


def test_quantized_numeric_ranges():
    db = xodb.inmemory()
    ctr = SchemaValueCountMatchSpy(AS, db)

    valnos = dict((a.name, db.add_value(a.name)) for a in (AS.bone, AS.wall, AS.seq))

    values = (10.0, 12.0, 13.0, 15.0, 16.0, 19.0, 23.0, 36.0)
    for i in range(len(values)):
        doc = Document()
        value = pickle.dumps(unicode(values[i]))
        doc.add_value(valnos['bone'], value)
        ctr(doc)

    assert ctr.build_numeric_ranges('bone', 0) == []
    assert ctr.build_numeric_ranges('bone', 2, 5.0) == [(10.0, 25.0), (25.0, 40.0)]
    assert ctr.build_numeric_ranges('bone', 3, 5.0) == [(10.0, 20.0), (20.0, 30.0), (30.0, 40.0)]
    assert ctr.build_numeric_ranges('bone', 7, 5.0) == [
        (10.0, 15.0), (15.0, 20.0), (20.0, 25.0), (25.0, 30.0), (30.0, 35.0), (35.0, 40.0)]

    assert ctr.build_numeric_ranges('bone', 2, 10.0) == [(10.0, 20.0), (20.0, 40.0)]


def test_date_ranges():
    db = xodb.inmemory()
    ctr = SchemaValueCountMatchSpy(AS, db)

    ranges = ctr.build_date_ranges(from_date=datetime.date(2009, 1, 1))
    assert ranges == [
             ('20090101', '20090101', 'today'),
             ('20081231', '20081231', 'yesterday'),
             ('20081230', '20081225', 'past week'),
             ('20090101', '20090131', 'January 2009'),
             ('20081201', '20081231', 'December 2008'),
             ('20081101', '20081130', 'November 2008'),
             ('20081001', '20081031', 'October 2008'),
             ('20080901', '20080930', 'September 2008'),
             ('20080801', '20080831', 'August 2008'),
             ('20080701', '20080731', 'July 2008'),
             ('20080601', '20080630', 'June 2008'),
             ('20080501', '20080531', 'May 2008'),
             ('20080401', '20080430', 'April 2008'),
             ('20080301', '20080331', 'March 2008'),
             ('20080201', '20080229', 'February 2008')]

    ranges = ctr.build_date_ranges(future=True, from_date=datetime.date(2009, 1, 1))
    assert ranges == [
             ('20090101', '20090101', 'today'),
             ('20090102', '20090102', 'tomorrow'),
             ('20090103', '20090108', 'this week'),
             ('20090101', '20090131', 'January 2009'),
             ('20090201', '20090228', 'February 2009'),
             ('20090301', '20090331', 'March 2009'),
             ('20090401', '20090430', 'April 2009'),
             ('20090501', '20090531', 'May 2009'),
             ('20090601', '20090630', 'June 2009'),
             ('20090701', '20090731', 'July 2009'),
             ('20090801', '20090831', 'August 2009'),
             ('20090901', '20090930', 'September 2009'),
             ('20091001', '20091031', 'October 2009'),
             ('20091101', '20091130', 'November 2009'),
             ('20091201', '20091231', 'December 2009')]




import datetime
import xodb
from xodb import Schema
from xodb.types import String, Numeric, Date, Datetime, Text, Sequence, Mapping

from nose.tools import assert_raises

# fixtures


class Thinger(object):

    def __init__(self, **args):
        self.__dict__ = args


class Stringer(Thinger):
    pass


class Numericer(Thinger):
    pass


class Dater(Thinger):
    pass


class Datetimer(Thinger):
    pass


class Texter(Thinger):
    pass


class Sequencer(Thinger):
    pass


class Mapper(Thinger):
    pass


class Person(Thinger):
    pass


class Department(Thinger):
    pass

# basic


class StringerSchema(Schema):
    s = String()


class NumericerSchema(Schema):
    n = Numeric()


class DaterSchema(Schema):
    d = Date()


class TexterSchema(Schema):

    def __before__(self, obj):
        self.__language__ = "en"
    
    t = Text(prefixed=False)
    j = Text()


class DatetimerSchema(Schema):
    dt = Datetime()


class SequencerSchema(Schema):
    seq = Sequence()


class MapperSchema(Schema):
    map = Mapping()

# sort


class SortStringerSchema(Schema):
    s = String(sortable=True)


class SortNumericerSchema(Schema):
    n = Numeric(sortable=True)


class SortDaterSchema(Schema):
    d = Date(sortable=True)


class SortDatetimerSchema(Schema):
    dt = Datetime(sortable=True)

# facets


class FacetStringerSchema(Schema):
    s = String(facet=True)


class FacetNumericerSchema(Schema):
    n = Numeric(facet=True)


class FacetDaterSchema(Schema):
    d = Date(facet=True)


class FacetDatetimerSchema(Schema):
    dt = Datetime(facet=True)


class FacetSequencerSchema(Schema):
    seq = Sequence(facet=True)


class FacetMapperSchema(Schema):
    map = Mapping(facet=True)

# containers


class SequencerNumericSchema(Schema):
    seq = Sequence(Numeric())


class SequencerSequencerSchema(Schema):
    seq = Sequence(Sequence())


class MapperNumericSchema(Schema):
    map = Mapping(Numeric())


class MapperMapperSchema(Schema):
    map = Mapping(Mapping())

#chaining


class ChainedStringerNumericerSchema(Schema):
    t = Text(getter=String(sortable=True))

    @Sequence.getter()
    @String.getter(sortable=True)
    def s(self, name):
        return getattr(self.__object__, name)

# modely stuff


class FacetPersonSchema(Schema):
    name = String(facet=True)


class FacetPersonGetterSchema(Schema):

    @String.getter(facet=True)
    def name(self, name):
        return getattr(self.__object__, name) + 'ha'


class FacetPersonRequiredSchema(Schema):
    zoo = String(facet=True, required=True)


class FacetPersonTypeRequiredSchema(Schema):
    name = String(facet=True, required_type=str)


class FacetDepartmentSchema(Schema):
    name = String(facet=True)
    employees = Sequence(facet=True)


class _TestXapianBase(object):
    db_factory = None

    def setup(self):
        self.db = self.db_factory()
        assert self.db.backend.get_doccount() == 0

    def test_string_term(self):
        self.db.map(Stringer, StringerSchema)
        s = Stringer(s="joe")
        self.db.add(s)
        assert self.db.backend.get_doccount() == 1
        assert self.db.query('s:joe').next().s == 'joe'

    def test_string_sort(self):
        self.db.map(Stringer, SortStringerSchema)
        s = Stringer(s="joe")
        s2 = Stringer(s="jane")
        self.db.add(s, s2)
        assert self.db.backend.get_doccount() == 2
        assert self.db.query('s:joe').next().s == 'joe'
        assert self.db.query('s:jane').next().s == 'jane'
        q = list(self.db.query('s:jane..joe', order='s'))
        assert [i.s for i in q] == ['jane', 'joe']
        q = list(self.db.query('s:jane..joe', order='s', reverse=True))
        assert [i.s for i in q] == ['joe', 'jane']

    def test_numeric_term(self):
        self.db.map(Numericer, NumericerSchema)
        n = Numericer(n=5)
        self.db.add(n)
        assert self.db.backend.get_doccount() == 1
        assert self.db.query('n:5').next().n == 5

    def test_numeric_sort(self):
        self.db.map(Numericer, SortNumericerSchema)
        n = Numericer(n=5)
        m = Numericer(n=6)
        self.db.add(n, m)
        assert self.db.backend.get_doccount() == 2
        assert self.db.query('n:5').next().n == 5
        assert self.db.query('n:6').next().n == 6
        q = list(self.db.query('n:4..8', order='n'))
        assert [i.n for i in q] == [5, 6]
        q = list(self.db.query('n:4..8', order='n', reverse=True))
        assert [i.n for i in q] == [6, 5]

    def test_date_term(self):
        self.db.map(Dater, DaterSchema)
        d = Dater(d=datetime.date(1975, 4, 23))
        self.db.add(d)
        assert self.db.backend.get_doccount() == 1
        d = datetime.date(1975, 4, 23)
        query = 'd:19750423'
        assert self.db.query('d:19750423').next().d == d
        for i in range(1, len(query) - 3):
            assert self.db.query(query[:-i] + '*').next().d == d

    def test_date_sort(self):
        self.db.map(Dater, SortDaterSchema)
        d = Dater(d=datetime.date(1975, 4, 23))
        d2 = Dater(d=datetime.date(1975, 5, 23))
        self.db.add(d, d2)
        assert self.db.backend.get_doccount() == 2
        assert (self.db.query('d:19750423').next().d ==
                datetime.date(1975, 4, 23))
        q = list(self.db.query('d:19750423..19750623', order="d"))
        assert len(q) == 2
        assert [i.d for i in q] == [d.d, d2.d]
        q = list(self.db.query('d:19750423..19750623',
                               order="d", reverse=True))
        assert len(q) == 2
        assert [i.d for i in q] == [d2.d, d.d]

    def test_datetime_term(self):
        self.db.map(Datetimer, DatetimerSchema)
        dt = Datetimer(dt=datetime.datetime(1975, 4, 23, 11, 59, 59))
        self.db.add(dt)
        assert self.db.backend.get_doccount() == 1
        dt = datetime.datetime(1975, 4, 23, 11, 59, 59)
        query = 'dt:19750423'
        assert self.db.query(query).next().dt == dt
        for i in range(1, len(query) - 3):
            assert self.db.query(query[:-i] + '*').next().dt == dt

    def test_datetime_sort(self):
        self.db.map(Datetimer, SortDatetimerSchema)
        dt = Datetimer(dt=datetime.datetime(1975, 4, 23, 11, 59, 59))
        dt2 = Datetimer(dt=datetime.datetime(1975, 5, 23, 11, 59, 59))
        self.db.add(dt, dt2)
        assert self.db.backend.get_doccount() == 2
        assert (self.db.query('dt:19750423').next().dt ==
                datetime.datetime(1975, 4, 23, 11, 59, 59))
        q = list(self.db.query('dt:19750423115959..19750623115959',
                               order="dt"))
        assert len(q) == 2
        assert [i.dt for i in q] == [dt.dt, dt2.dt]
        q = list(self.db.query('dt:19750423115959..19750623115959',
                               order="dt", reverse=True))
        assert len(q) == 2
        assert [i.dt for i in q] == [dt2.dt, dt.dt]

    def test_sequence_term(self):
        self.db.map(Sequencer, SequencerSchema)
        data = ['one', 'two', 'three']
        seq = Sequencer(seq=data)
        self.db.add(seq)
        assert self.db.backend.get_doccount() == 1
        for s in data:
            assert self.db.query('seq:%s' % s).next().seq == data

    def test_sequence_numeric_term(self):
        self.db.map(Sequencer, SequencerNumericSchema)
        data = [1, 2, 3]
        seq = Sequencer(seq=data)
        self.db.add(seq)
        assert self.db.backend.get_doccount() == 1
        for s in data:
            assert self.db.query('seq:%s' % s).next().seq == data

    def test_sequence_sequence_term(self):
        self.db.map(Sequencer, SequencerSequencerSchema)
        data = [['one', 'two', 'tree'],
                ['fee', 'fi', 'foo'],
                ['dee', 'doo', 'dum']]
        seq = Sequencer(seq=data)
        self.db.add(seq)
        assert self.db.backend.get_doccount() == 1
        for s in data:
            for j in s:
                assert self.db.query('seq:%s' % j).next().seq == data

    def test_mapping_term(self):
        self.db.map(Mapper, MapperSchema)
        data = {'one': 'ein',
                'two': 'zwei',
                'three': 'drei'}
        map = Mapper(map=data)
        self.db.add(map)
        assert self.db.backend.get_doccount() == 1
        for k, v in data.items():
            assert self.db.query('map_%s:%s' % (k, v)).next().map == data

    def test_mapping_numeric_term(self):
        self.db.map(Mapper, MapperNumericSchema)
        data = {'one': 1,
                'two': 2,
                'three': 3}
        map = Mapper(map=data)
        self.db.add(map)
        assert self.db.backend.get_doccount() == 1
        for k, v in data.items():
            assert self.db.query('map_%s:%s' % (k, v)).next().map == data

    def test_mapping_mapping_term(self):
        self.db.map(Mapper, MapperMapperSchema)
        data = {'one': dict(foo='fum', fin='fom'),
                'two': dict(fee='fi', fuu='fe'),
                'three': dict(dee='dum', duh='der')}
        map = Mapper(map=data)
        self.db.add(map)
        assert self.db.backend.get_doccount() == 1
        for ok, ov in data.items():
            for k, v in ov.items():
                assert self.db.query('map_%s_%s:%s' %
                                     (ok, k, v)).next().map == data

    def test_string_facet(self):
        self.db.map(Stringer, FacetStringerSchema)
        s = Stringer(s="joe")
        self.db.add(s)
        assert self.db.backend.get_doccount() == 1
        assert self.db.query('facet:s').next().s == 'joe'

    def test_numeric_facet(self):
        self.db.map(Numericer, FacetNumericerSchema)
        n = Numericer(n=5)
        self.db.add(n)
        assert self.db.backend.get_doccount() == 1
        assert self.db.query('facet:n').next().n == 5

    def test_text_term(self):
        self.db.map(Texter, TexterSchema)
        s = Texter(t="Joe's House of Running", j="the short and the prefixed.")
        self.db.add(s)
        assert self.db.backend.get_doccount() == 1
        assert (self.db.query('joe', lang="en").next().t
                == "Joe's House of Running")
        assert (self.db.query('house', lang="en").next().t
                == "Joe's House of Running")
        assert (self.db.query('run', lang="en").next().t
                == "Joe's House of Running")
        assert not set(self.db.query('prefix', lang="en"))

        assert (self.db.query('j:short', lang="en").next().j
                == "the short and the prefixed.")
        assert (self.db.query('j:prefix', lang="en").next().j
                == "the short and the prefixed.")
        assert not set(self.db.query('j:house', lang="en"))

    def test_date_facet(self):
        self.db.map(Dater, FacetDaterSchema)
        d = Dater(d=datetime.date(1975, 4, 23))
        self.db.add(d)
        assert self.db.backend.get_doccount() == 1
        assert self.db.query('facet:d').next().d == d.d

    def test_datetime_facet(self):
        self.db.map(Datetimer, FacetDatetimerSchema)
        dt = Datetimer(dt=datetime.datetime(1975, 4, 23))
        self.db.add(dt)
        assert self.db.backend.get_doccount() == 1
        assert self.db.query('facet:dt').next().dt == dt.dt

    def test_sequence_facet(self):
        self.db.map(Sequencer, FacetSequencerSchema)
        data = ['one', 'two', 'three']
        seq = Sequencer(seq=data)
        self.db.add(seq)
        assert self.db.backend.get_doccount() == 1
        assert self.db.query('facet:seq').next().seq == data

    def test_mapping_facet(self):
        self.db.map(Mapper, FacetMapperSchema)
        data = {'one': 1,
                'two': 2,
                'three': 3}
        map = Mapper(map=data)
        self.db.add(map)
        assert self.db.backend.get_doccount() == 1
        for k in data:
            assert self.db.query('facet:map_' + k).next().map == data

    def test_suggest_facet(self):
        self.db.map(Person, FacetPersonSchema)
        self.db.map(Department, FacetDepartmentSchema)
        o = Person(name="joe")
        a = Person(name="jane")
        d = Department(name="housing", employees=[o.name, a.name])
        self.db.add(o, a, d)

        assert self.db.backend.get_doccount() == 3
        assert (set(self.db.suggest('name:housing', prefix='facet')) ==
                set([u'facet:employees', u'facet:name']))
        assert (set(self.db.suggest('name:housing OR name:jane',
                                    prefix='facet')) ==
                set([u'facet:employees', u'facet:name']))
        assert not set(self.db.suggest('name:housing AND name:jane',
                                       prefix='facet'))

        assert (set(self.db.suggest('name:joe', prefix='facet')) ==
                set([u'facet:name']))
        assert (set(self.db.suggest('name:jane', prefix='facet')) ==
                set([u'facet:name']))
        assert (set(self.db.suggest('name:jane OR name:joe',
                                    prefix='facet')) ==
                set([u'facet:name']))
        assert not set(self.db.suggest('name:jane AND name:joe',
                                       prefix='facet'))

        assert (set(self.db.suggest('employees:jane', prefix='facet')) ==
                set([u'facet:employees', u'facet:name']))
        assert (set(self.db.suggest('employees:joe', prefix='facet')) ==
                set([u'facet:employees', u'facet:name']))
        assert (set(self.db.suggest('employees:joe OR employees:jane',
                                    prefix='facet')) ==
                set([u'facet:employees', u'facet:name']))
        assert (set(self.db.suggest('employees:joe AND employees:jane',
                                    prefix='facet')) ==
                set([u'facet:employees', u'facet:name']))
        assert not set(self.db.suggest('employees:joe AND name:jane',
                                       prefix='facet'))

    def test_getter(self):
        self.db.map(Person, FacetPersonGetterSchema)
        self.db.map(Department, FacetDepartmentSchema)
        o = Person(name="joe")
        a = Person(name="jane")
        self.db.add(o, a)
        assert self.db.query("name:joeha").next().name == 'joeha'

    def test_required(self):
        self.db.map(Person, FacetPersonRequiredSchema)
        o = Person(name="joe")
        assert_raises(xodb.exc.AttributeRequired, self.db.add, o)

    def test_type_required(self):
        self.db.map(Person, FacetPersonTypeRequiredSchema)
        o = Person(name=4)
        assert_raises(xodb.exc.AttributeTypeRequired, self.db.add, o)

    def test_chained_string_text(self):
        self.db.map(Stringer, ChainedStringerNumericerSchema)
        s = Stringer(s='abc', t='once upon a lamb')
        self.db.add(s)
        assert self.db.query("s:abc").next().s == 'abc'
        assert self.db.query("s:a").next().s == 'abc'
        assert self.db.query("s:b").next().s == 'abc'
        assert self.db.query("s:c").next().s == 'abc'
        assert self.db.query("t:once").next().t == 'once upon a lamb'
        doc = self.db.query("t:once", document=True).next()
        assert doc.get_value(self.db.values['s']) == 'abc'
        assert doc.get_value(self.db.values['t']) == 'once upon a lamb'


class TestXapianFile(_TestXapianBase):
    db_factory = staticmethod(xodb.temp)


class TestXapianInMem(_TestXapianBase):
    db_factory = staticmethod(xodb.inmemory)



from __future__ import absolute_import

import xapian

from . import snowball
from . util import lazy_property
from . fields import Text, PrefixedText, Term, PrefixedTerm, Value


_default_marker = object()


class Attribute(object):

    term_field = None
    prefixed_field = None
    value_field = None

    def __init__(self, type=None, required=True,
                 getter=None, volatile=False,
                 prefixed=False, unprefixed=True,
                 value=False, text=False,
                 value_parent=None, sequence=False,
                 mapping=False, sortable=None,
                 translit=None, default=_default_marker):
        self.name = None # set by the schema construction
        self.type = type
        self.required = required
        self.default = default
        self.getter = getter
        self.volatile = volatile
        self.children = []
        self.value_parent = value_parent
        self.sequence = sequence
        self.mapping = mapping
        self.sortable = sortable
        if text:
            if unprefixed:
                self.term_field = Text(self, translit=translit)
            if prefixed:
                self.prefixed_field = PrefixedText(self, prefix=self.name,
                                                   translit=translit)
        else:
            if unprefixed:
                self.term_field = Term(self)
            if prefixed:
                self.prefixed_field = PrefixedTerm(self, prefix=self.name)
        if value or sortable:
            self.value_field = Value(self, parent=value_parent,
                                     sortable=sortable,
                                     prefix=self.name)

    def __call__(self, schema, name, value=None):
        obj = schema.__object__
        state = schema.__state__

        if self.volatile:
            if name in state:
                del state[name]
            return

        if self.getter is not None:
            value = self.getter(name, schema)

        elif name in state:
            value = state[name]

        elif self.default is not _default_marker:
            value = self.default

        elif self.required:
            msg = "Attribute %s is required for %s by %s" % (name,
                                                             repr(obj),
                                                             self.__class__)
            raise TypeError(msg)

        if self.type is not None:
            if not isinstance(value, self.type):
                msg = "Attribute %s must be of type %s" % (name,
                                                           self.type)
                raise TypeError(msg)

        state[name] = value

        for field in [self.term_field, self.prefixed_field, self.value_field]:
            if field:
                field(schema, field.attr.name, obj, state)


class Schema(object):
    """
    Types that define indexing behavior for objects.  Instances of
    these types are used at index-time to add terms, values, and data
    to Xapian documents.
    """
    __slots__ = ['__database__',
                 '__document__',
                 '__object__',
                 '__state__',
                 '__term_generator__',
                 '__language__',
                 ]

    class __metaclass__(type):
        def __new__(cls, cls_name, bases, attrs):
            schema = type.__new__(cls, cls_name, bases, attrs)
            schema.__attributes__ = {}
            for base in reversed(schema.__mro__):
                if base is not schema:
                    if hasattr(base, '__attributes__'):
                        schema.__attributes__.update(base.__attributes__)
            for name, value in attrs.iteritems():
                if not name.startswith("_"):
                    value.name = name
                    schema.__attributes__[name] = value
            return schema

    def __init__(self, database):
        self.__database__ = database
        self.__language__ = None

    def __call__(self, obj):
        if hasattr(self, '__before__'):
            self.__before__(obj)
        state = newargs = None
        if hasattr(obj, '__getstate__'):
            state = obj.__getstate__()
        else:
            state = obj.__dict__.copy()

        for key in state.keys():
            if key not in self.__attributes__:
                state.pop(key)

        if hasattr(obj, '__getnewargs__'):
            newargs = obj.__getnewargs__()
        else:
            newargs = ()

        self.__object__ = obj
        self.__state__ = state
        typ = type(obj)

        self.__document__ = doc = xapian.Document()
        if self.__attributes__:
            self.__term_generator__ = tg = xapian.TermGenerator()
            tg.set_database(self.__database__.backend)
            tg.set_document(doc)
            if self.__database__.spelling:
                tg.set_flags(xapian.TermGenerator.FLAG_SPELLING)
            language = self.__language__
            if language in snowball.stoppers:
                tg.set_stemmer(xapian.Stem(language))
                tg.set_stopper(snowball.stoppers[language])

            for name, attr in self.__attributes__.iteritems():
                attr(self, name)

        data = self.__database__.serialize(typ, newargs, state)
        doc.set_data(data)
        if hasattr(self, '__after__'):
            self.__after__(obj, state, doc)
        return doc



import datetime
import weakref
import gc

from xodb import Schema, Attribute

from nose.tools import assert_raises


class Foo(object): 
    def __init__(self, bar=None):
        self.bar = bar


class Bar(object):
    def __init__(self, ding=None):
        self.ding = ding

    def __len__(self):
        return 42


class Ding(object):
    def __init__(self, bar):
        self.bar = bar
        self.blig = 41
        self.bloog = 42
        self.sez = set([])
        self.canary = []

        self.language = 'en'
        self.name = 'ding'
        self.direction = 'north'
        self.title = 'The young and the database.'
        self.instructions = 'Under the walker'
        self.zig = 3
        self.zag = 6
        self.zork = ['zap', 'zop', 'zern']
        self.zirk = {'jib': 'zoop', 'jab': 'zeep'}
        self.bing = ['boop', 'berp', 'barp']
        self.bang = {'plop': 'This is a borp of a time', 'plip': 'Textual birpism'}
        self.slap = 3
        self.slag = datetime.date(2009, 01, 20)
        self.slat = 'fuzzy'


class DingSchema(Schema):
    bar  = Attribute(type=Bar)
    zig  = Attribute(required=False)
    blig = Attribute(volatile=True)


class Bling(Foo):
    def __new__(cls, blargh):
        inst = super(Bling, cls).__new__(cls)
        inst.blargh = blargh
        return inst

    def __getstate__(self):
        return dict(value='fuzzy')

    def __setstate__(self, state):
        assert state['value'] == 'fuzzy'
        setattr(self, 'bob', 'dinge')

    def __getnewargs__(self):
        return ('yo',)


class FongSchema(Schema):
    pass


class Fern(object):
    __xodb_schema__ = FongSchema


class _BaseTest(object):
    ding_schema = DingSchema

    fong_schema = FongSchema

    def test_bogus_oid(self):
        s = self.db.new_session()
        assert_raises(KeyError, s.get_oid, "foo")

    def test_unmap(self):
        self.db.mapper.unmap(int)
        assert int not in self.db.mapper.type_map

    def test_clear_mapper(self):
        self.db.mapper.clear()
        assert not self.db.mapper.type_map

    def test_add_and_commit(self):
        """Test adding an object and commiting the transaction.. """
        f = Foo()
        s = self.db.new_session()
        s.add(f)
        s.commit()
        assert len(self.db) == 1
        assert f in s

    def test_delete_and_commit(self):
        f = Foo()
        s = self.db.new_session()
        s.add(f)
        s.commit()
        oid = s.get_oid(f)
        assert len(self.db) == 1
        del self.db.backend[oid]
        assert len(self.db) == 0

    def test_add_and_cancel(self):
        """Test adding an object and cancelling the transaction.. """
        f = Foo()
        s = self.db.new_session()
        s.add(f)
        s.cancel()
        assert len(self.db) == 0
        assert f in s

    def test_query_cache_clear_and_gc(self):
        """Test session, make sure cleared objects do not leak. """
        f = Foo()
        s = self.db.new_session()
        s.add(f)
        s.commit()
        foid = s.get_oid(f)
        assert f in s

        results = list(s.query(foid))
        assert len(results) == 1
        assert results[0] is f

        # make sure when a session is cleared that the objects are not
        # leaked
        d = weakref.WeakValueDictionary()
        d['f'] = f
        s.clear()
        del f, results
        gc.collect()

        # the dict is empty, the object is gone
        assert len(d) == 0

        # requery for it
        results = list(s.query(foid))
        assert len(results) == 1
        assert type(results[0]) is Foo

    def test_cache_reference_and_clear(self):
        """ Test reference resolving, before and after session clear. """
        b = Bar()
        f = Foo(b)
        # f, a Foo, now references b, a Bar

        # let's add it to a db
        s = self.db.new_session()
        s.add(f)
        s.commit()

        # there are two documents, let's get their ids
        assert len(self.db) == 2
        foid = s.get_oid(f)
        boid = s.get_oid(b)

        # get f
        results = list(s.query(foid))
        assert len(results) == 1

        # assert it was cached in session, it's the same object
        assert results[0] is f

        # get b
        results = list(s.query(boid))
        assert len(results) == 1

        # assert it was cached in session
        assert results[0] is b

        # now let's clear this session
        s.clear()

        # get f again, check it's type
        results = list(s.query(foid))
        assert len(results) == 1
        f2 = results[0]
        assert type(f2) is Foo

        # since we cleared the session, the new object is not the same as
        # before
        assert f2 is not f

        # get b again, same story with it
        results = list(s.query(boid))
        assert len(results) == 1
        b2 = results[0]
        assert type(b2) is Bar
        assert b2 is not b

        # f2.bar is a ghost, let's make a local reference to it so we
        # don't loose it.
        d = f2.bar
        # make sure it really is a ghost
        assert type(d) is Ghost
        # ghosts are not in a session
        assert d not in s
        # taking the len loads the real object into f2 (doesn't change d!)
        assert len(f2.bar) == 42
        # now the ghost is not longer f2.bar (still d though)
        assert type(f2.bar) is Bar
        # bar can be traversed
        assert f2.bar.ding is None
        # the object hasn't changed
        assert b2 is f2.bar

        # now let's try to setattr on the ghost
        d.zig = 3
        # make sure it's set
        assert d.zig == 3
        # on the real object too
        assert f2.bar.zig == 3
        # make sure we can del it
        del d.zig
        assert not hasattr(d, 'zig')
        # on the real object too
        assert not hasattr(f2.bar, 'zig')
        # make sure we're still working with a ghost after all that
        assert type(d) is Ghost

    def test_required_attribute(self):
        """ Test an error is raised in a required schema attribute is missing. """
        s = self.db.new_session()
        self.db.mapper.map(Ding, self.ding_schema)
        d = Ding(Bar())
        del d.bar
        s.add(d)
        assert_raises(TypeError, s.commit)

    def test_volatile_attribute(self):
        s = self.db.new_session()
        self.db.mapper.map(Ding, self.ding_schema)
        d = Ding(Bar())
        assert hasattr(d, 'blig')
        s.add(d)
        s.commit()
        doid = s.get_oid(d)
        results = list(s.query(doid))
        assert results[0] is d
        s.clear()
        results = list(s.query(doid))
        assert not hasattr(results[0], 'blig')

    def test_changed_object(self):
        d = Ding(Bar())
        self.db.mapper.map(Ding, self.ding_schema)
        s = self.db.new_session()
        s.add(d)
        s.commit()
        doid = s.get_oid(d)
        s.clear()
        d = s.query(doid).next()

        d.blig = 1
        d.bloog = 1
        s.commit()
        s.clear()
        d = s.query(doid).next()
        assert not hasattr(d, 'blig')
        assert d.bloog == 42

        d.blig = 1
        d.bloog = 1
        s.changed(d)
        d.canary = [] # ewww
        s.commit()
        s.clear()
        d = s.query(doid).next()
        assert not hasattr(d, 'blig')
        assert d.bloog == 1

    def test_pickle_protocol(self):
        """Test __getstate__, __setstate__, and __getnewargs__ """
        d = Bling('flerg')
        assert d.blargh == 'flerg'
        s = self.db.new_session()
        s.add(d)
        s.commit()
        doid = s.get_oid(d)
        s.clear()
        d = s.query(doid).next()
        assert d.bob == 'dinge'
        assert d.blargh == 'yo'

    def test_persistent_add_and_commit(self):
        f = Foop(a="42", b=42, c=Foop(z=1))
        assert f._p_status is persistent.UNSAVED
        s = self.db.new_session()
        s.add(f)
        s.commit()
        assert len(self.db) == 2
        assert f in s
        assert f._p_status is persistent.SAVED
        foid = s.get_oid(f)
        s.clear()
        f = s.query(foid).next()
        assert f._p_status is persistent.GHOST
        assert f.a == "42"
        assert f._p_status is persistent.SAVED

        assert f.c._p_status is persistent.GHOST
        assert f.c.z == 1
        assert f.c._p_status is persistent.SAVED

    def test_persistent_changed(self):
        f = Foop(a="42", b=42, c=Foop(z=1))
        s = self.db.new_session()
        s.add(f)
        s.commit()
        foid = s.get_oid(f)
        s.clear()
        f = s.query(foid).next()
        assert f._p_status is persistent.GHOST
        del f.a
        assert f._p_status is persistent.UNSAVED
        assert not hasattr(f, 'a')
        s.commit()
        s.clear()
        f = s.query(foid).next()
        assert f._p_status is persistent.GHOST
        assert not hasattr(f, 'a')
        assert f._p_status is persistent.SAVED
        s.clear()
        f = s.query(foid).next()
        assert f._p_status is persistent.GHOST
        f.a = 43
        assert f._p_status is persistent.UNSAVED
        s.commit()
        assert f._p_status is persistent.SAVED
        s.clear()
        f = s.query(foid).next()
        assert f._p_status is persistent.GHOST
        assert hasattr(f, 'a')
        assert f.a == 43
        assert f._p_status is persistent.SAVED

    def test_declared_schema(self):
        d = Ding(Bar())
        self.db.mapper.map(Ding, self.declared_ding_schema)
        s = self.db.new_session()
        s.add(d)
        s.commit()
        doid = s.get_oid(d)
        s.clear()
        d = s.query(doid).next()
        assert hasattr(d, 'bar')
        assert type(d.bar) is Ghost
        assert type(d.bloog) is int
        assert d.bloog == 42
        assert not hasattr(d, 'blig')
        
    def test_explicit_class_schema(self):
        canary = []
        fong_schema = self.fong_schema
        def monkey(self, session):
            canary.append(session)
            super(fong_schema, self).__init__(session)
        fong_schema.__init__ = monkey
        Fern.__xodb_schema__ = fong_schema
        f = Fern()
        s = self.db.new_session()
        s.add(f)
        s.commit()
        assert canary[0] is s



from __future__ import absolute_import

import tempfile
import xapian

from . import exc
from . database import Database, Record, MultipleValueRangeProcessor, instance_factory
from . schema import Schema, Attribute
from . types import String, Numeric, Text, Sequence, Mapping, Date, Datetime


def open(path_or_db, writable=True, overwrite=False, spelling=True, record_factory=Record):
    """Return an xodb database with the given path or xapian database object.

    :param path_or_db: A path to a database file or a pre-existing xapian database object.

    :param writable: Open database in writable mode.

    :param overwrite: If writable is True, overwrite the existing
    database with a new one.

    :param spelling: If True, write spelling correction data to the
    database.

    :param record_factory: Callable factory to pass stored object
    state when results are instanciated.
    """
    return Database(path_or_db, writable=writable, overwrite=overwrite, 
                    spelling=spelling, record_factory=record_factory)


def temp(record_factory=Record, spelling=True):
    """Returns an xodb database backed by a teporary directory.  You
    are responsible for cleaning up the directory.

    :param spelling: If True, write spelling correction data to database.

    :param record_factory: Callable factory to pass stored object
    state when results are instanciated.
    """ 
    return open(tempfile.mkdtemp(), spelling=spelling, record_factory=record_factory)


def inmemory(record_factory=Record):
    """Returns an xodb database backed by an in-memory xapian
    database.  Does not support spelling correction.

    :param record_factory: Callable factory to pass stored object
    state when results are instanciated.
    """
    return open(xapian.inmemory_open(), spelling=False, record_factory=record_factory)



import xodb
from xodb import Schema, Attribute


class A(object):
    def __init__(self, **kw):
        for k,v in kw.items():
            setattr(self, k, v)


class B(A):
    pass


class AS(Schema):
    id = Attribute()
    bone = Attribute(required=False)
    slow = Attribute()


class BS(AS):
    fod = Attribute()
    bone = Attribute()


def test_basic_inheritance():
    assert issubclass(BS, AS)
    assert not issubclass(AS, BS)
    assert AS.__attributes__['bone'].required is False
    assert AS.bone.required is False
    assert BS.__attributes__['bone'].required is True
    assert BS.bone.required is True
    assert 'id' in BS.__attributes__
    assert hasattr(BS, 'id')
    assert 'fod' not in AS.__attributes__
    assert not hasattr(AS, 'fod')




import xapian
import unicodedata

from .exc import (
    AttributeRequired,
    AttributeTypeRequired,
    InvalidTermError,
    )


MAX_TERM_LEN = 240 # compile time xapian limit


def _prefix(name):
    return (u'X' + name.upper()).encode('utf-8')


_default_marker = object()


class BaseType(object):
    """
    A schema type.

    :param sort: If true, this attribute is sortable and will create a
    xapian value.

    :param facet: If true, this attribute is faceted and will generate
    facet terms.

    :param facet_parent: If facet is true, this attributes facet
    parent.  Used to organize facet heirarchies (ie country-> state->
    city etc.)

    :param prefixed: If true, this attribute's value should be prefixed.

    :param getter: A getter function for the value.  If None, the
    value is taken from the object's state.

    :param required: If true, the object is required to provide this
    attribute.

    :param default: The default value for the attribute if one is not
    provided.  Meaningless if required is True.

    :param required_type: If not None, the value of the attribute must
    be an instance of this type.
    """

    def __init__(self, getter=None, sortable=False,
                 facet=False, facet_parent=None, prefixed=True,
                 required=False, translit=None,
                 default=_default_marker, required_type=None):
        self.sortable = sortable
        self.prefixed = prefixed
        self.facet = facet
        self.facet_parent = facet_parent
        self.default = default
        self.getter = getter
        self.required = required
        self.translit = translit
        self.required_type = required_type
        self.valno = None

    @classmethod
    def getter(type_class, *args, **kwargs):
        """
        Decorator for wrapping a schema getter method with a given
        schema type.
        """
        def _decorator(f):
            kwargs['getter'] = f
            t = type_class(*args, **kwargs)
            return t
        return _decorator

    def normalize(self, value, limit=MAX_TERM_LEN):
        value = unicodedata.normalize(
            'NFKC', unicode(value).strip().lower()).encode('utf-8')
        if limit and len(value) > limit:
            raise InvalidTermError('The term %s is too long.' % value)
        return value

    def check_prefix(self, schema, name, boolean=False):
        database = schema.__database__
        prefixes = database.relevance_prefixes.copy()
        prefixes.update(database.boolean_prefixes.copy())
        if name not in prefixes:
            upped = _prefix(name)
            if boolean:
                database.add_boolean_prefix(name, upped)
            else:
                database.add_prefix(name, upped)

    def add_term(self, schema, value, prefix=None, boolean=False):
        value = self.normalize(value)
        if prefix:
            self.check_prefix(schema, prefix, boolean)
            term = _prefix(prefix) + value.lower()
        else:
            term = value.lower()
        schema.__document__.add_term(term)

    def add_value(self, schema, name, value):
        database = schema.__database__
        if name in database.values:
            valno = database.values[name]
        else:
            valno = database.add_value(name, self.__class__.__name__.lower())
        self.valno = valno
        schema.__document__.add_value(valno, value)

    def __call__(self, schema, name, value=None):
        state = schema.__state__
        obj = schema.__object__
        if value is None:
            if self.getter is not None:
                value = self.getter(schema, name)

            elif name in state:
                value = state[name]

            elif self.default is not _default_marker:
                value = self.default

            elif self.required:
                msg = "Attribute %s is required for %s by %s" % (name,
                                                                 repr(obj),
                                                                 self.__class__)
                raise AttributeRequired, msg
            if self.required_type is not None:
                if not isinstance(value, self.required_type):
                    msg = "Attribute %s must be of type %s" % (name,
                                                               self.required_type)
                    raise AttributeTypeRequired, msg
            if self.translit:
                value = value.encode(self.translit)
            if self.facet:
                self.add_term(schema, name, 'facet')
            state[name] = value
        return value


class String(BaseType):
    """A schema type that indexes a string attribute.
    """
    def __call__(self, schema, name, value=None):
        value = BaseType.__call__(self, schema, name, value)
        if value:
            if self.prefixed:
                self.add_term(schema, value, name)
            else:
                self.add_term(schema, value)
            if self.sortable:
                self.add_value(schema, name, value)
        return value


class Numeric(BaseType):
    """A schema type that indexes a numeric attribute.
    """
    def __call__(self, schema, name, value=None):
        value = BaseType.__call__(self, schema, name, value)
        if value:
            if self.prefixed:
                self.add_term(schema, value, name)
            else:
                self.add_term(schema, value)
            if self.sortable:
                self.add_value(schema, name, xapian.sortable_serialise(value))
        return value


class Text(BaseType):
    """A schema type that indexes a textual attribute.
    """
    def __init__(self, getter=None, positions=True,
                 *args, **kwargs):
        BaseType.__init__(self, getter, *args, **kwargs)
        self.positions = positions

    def __call__(self, schema, name, value=None):
        value = BaseType.__call__(self, schema, name, value)
        if value:
            tg = schema.__term_generator__
            if self.positions:
                index_text = tg.index_text
            else:
                index_text = tg.index_text_without_positions
            if self.prefixed:
                self.check_prefix(schema, name)
                index_text(value.encode('utf-8'), 1, _prefix(name))
            else:
                index_text(value.encode('utf-8'))
        return value


class Sequence(BaseType):
    def __init__(self, itemizer=None, *args, **kwargs):
        BaseType.__init__(self, *args, **kwargs)
        self.itemizer = itemizer
    
    def __call__(self, schema, name, value=None):
        value = BaseType.__call__(self, schema, name, value)
        if value:
            for val in value:
                if self.itemizer is not None:
                    self.itemizer(schema, name, val)
                else:
                    if self.prefixed:
                        self.add_term(schema, val, name)
                    else:
                        self.add_term(schema, val)
        return value


class Mapping(BaseType):
    def __init__(self, itemizer=None, *args, **kwargs):
        BaseType.__init__(self, *args, **kwargs)
        self.itemizer = itemizer

    def __call__(self, schema, name, value=None):
        value = BaseType.__call__(self, schema, name, value)
        if value:
            for k, v in value.iteritems():
                if self.itemizer is not None:
                    self.itemizer(schema, (name + '_' + k), v)
                else:
                    if self.prefixed:
                        n = '%s_%s' % (name, k)
                    else:
                        n = k
                    self.add_term(schema, v, n)
                    if self.facet:
                        self.add_term(schema, n, 'facet')
        return value


class BaseDateType(BaseType):
    """A schema type that indexes a mapping attribute.
    """
    def __call__(self, schema, name, value=None):
        value = BaseType.__call__(self, schema, name, value)
        if value:
            term_value = value.strftime(self.term_format)
            if self.prefixed:
                self.add_term(schema, term_value, name)
            else:
                self.add_term(schema, term_value)
            if self.sortable:
                value_value = value.strftime(self.value_format)
                self.add_value(schema, name, value_value)
        return value


class Date(BaseDateType):
    """A schema type that indexes a date attribute.
    """
    term_format = '%Y%m%d'
    value_format = '%Y%m%d'


class Datetime(BaseDateType):
    """A schema type that indexes a datetime attribute.
    """
    term_format = '%Y%m%d'
    value_format = '%Y%m%d%H%M%S'



import xapian
import unicodedata
import translitcodec
from cPickle import dumps, loads

from . import snowball

MAX_TERM_LEN = 240 # compile time xapian limit


def _prefix(name):
    return (u'X'+ name.upper()).encode('utf-8')


_novalue = object()


class Field(object):
    __slots__ = ['attr', 'prefix']

    def __init__(self, attr, prefix=None):
        self.attr = attr
        self.prefix = prefix

    def check_prefix(self, schema, prefix, boolean=False):
        """
        Check the database in the current session for the given
        prefix.  If it doesn't exist, add it.
        """
        database = schema.__database__
        prefixes = database.relevance_prefixes.copy()
        prefixes.update(database.boolean_prefixes.copy())
        if prefix not in prefixes:
            upped = _prefix(prefix)
            if boolean:
                database.add_boolean_prefix(prefix, upped)
            else:
                database.add_prefix(prefix, upped)

    def __call__(self, schema, name, obj, state):
        self.check_prefix(schema, (self.prefix or self.attr.name))
        self.index(schema, name, obj, state)

    def normalize(self, value):
        value = unicodedata.normalize('NFKC', unicode(value).strip().lower()).encode('utf-8')
        if len(value) > MAX_TERM_LEN:
            return ''
        return value

    def add_term(self, schema, value, prefix=None):
        value = self.normalize(value)
        if value:
            if prefix is None:
                schema.__document__.add_term(value)
            else:
                term = _prefix(prefix) + value.lower()
                schema.__document__.add_term(term)

    def index(self, schema, name, obj, state):
        raise NotImplementedError


class Term(Field):
    """Field that adds a single term into a schema document. """

    def index(self, schema, name, obj, state):
        value = state.get(self.attr.name)
        if value:
            prefix = self.prefix or self.attr.name
            if self.attr.sequence:
                for val in value:
                    self.add_term(schema, val)
            elif self.attr.mapping:
                for key, val in value.items():
                    if not val:
                        continue
                    pre = prefix + "_" + key
                    self.check_prefix(schema, pre)
                    self.add_term(schema, val, prefix=pre)
            else:
                self.add_term(schema, value)


class PrefixedTerm(Field):
    """Field that adds a single prefixed term into a schema document."""

    def index(self, schema, name, obj, state):
        value = state.get(self.attr.name)
        if value:
            prefix = self.prefix or self.attr.name
            if self.attr.sequence:
                for val in value:
                    self.add_term(schema, val, prefix=prefix)
            elif self.attr.mapping:
                for key, val in value.items():
                    if not val:
                        continue
                    pre = prefix + "_" + key
                    self.check_prefix(schema, pre)
                    self.add_term(schema, val, prefix=pre)
            else:
                self.add_term(schema, value, prefix=prefix)


class Text(Field):
    """A field that indexes text into a schema document."""

    __slots__ = ['translit']

    def __init__(self, attr, translit=None):
        Field.__init__(self, attr)
        self.translit = translit

    def index(self, schema, name, obj, state):
        value = state.get(self.attr.name)
        if value:
            tg = schema.__term_generator__
            language = schema.__language__
            if language in snowball.stoppers:
                tg.set_stemmer(xapian.Stem(language))
                tg.set_stopper(snowball.stoppers[language])
            prefix = self.prefix or self.attr.name
            if self.attr.sequence:
                for val in value:
                    if not val:
                        continue
                    if self.translit:
                        val = val.encode(self.translit)
                    tg.index_text(val.encode('utf-8'))
            elif self.attr.mapping:
                for key, val in value.items():
                    if not val:
                        continue
                    if self.translit:
                        val = val.encode(self.translit)
                    pre = prefix + "_" + key
                    self.check_prefix(schema, pre)
                    tg.index_text(val.encode('utf-8'), 1, _prefix(pre))
            else:
                if self.translit:
                    value = value.encode(self.translit)
                tg.index_text(value.encode('utf-8'))


class PrefixedText(Field):
    """A field that indexes prefixed text into a schema document."""

    __slots__ = ['translit']

    def __init__(self, attr, prefix=None, translit=None):
        Field.__init__(self, attr, prefix)
        self.translit = translit

    def index(self, schema, name, obj, state):
        value = state.get(self.attr.name)
        if value:
            prefix = self.prefix or self.attr.name
            tg = schema.__term_generator__
            language = schema.__language__
            if language in snowball.stoppers:
                tg.set_stemmer(xapian.Stem(language))
                tg.set_stopper(snowball.stoppers[language])
            if self.attr.sequence:
                for val in value:
                    if not val:
                        continue
                    if self.translit:
                        val = val.encode(self.translit)
                    tg.index_text(val.encode('utf-8'), 1, _prefix(prefix))
            elif self.attr.mapping:
                for key, val in value.items():
                    if not val:
                        continue
                    if self.translit:
                        val = val.encode(self.translit)
                    pre = prefix + "_" + key
                    self.check_prefix(schema, pre)
                    tg.index_text(val.encode('utf-8'), 1, _prefix(pre))
            else:
                if self.translit:
                    value = value.encode(self.translit)
                tg.index_text(value.encode('utf-8'), 1, _prefix(prefix))


class Value(Field):
    """A document value used for matching and sorting."""

    __slots__ = ['attr', 'filter', 'valno', 'parent', 'sortable']

    def __init__(self, attr, parent=None, filter=None, sortable=None, prefix=None):
        Field.__init__(self, attr, prefix)
        self.attr = attr
        self.filter = filter
        self.parent = parent

        if sortable not in (None, 'numeric', 'date', 'datetime', 'string'):
            raise TypeError("Unkown sortable type: %s" % sortable)
        self.sortable = sortable
        if parent:
            # sw: attr.name is net yet known here
            parent.children.append(attr)

    def __call__(self, schema, name, obj, state):
        self.check_value(schema)
        if self.filter:
            self.attr.value = self.filter(self.attr.value)
        self.index(schema, name, obj, state)

    def check_value(self, schema):
        """Check the database in the current session for the given
        value name.  If it doesn't exist, add it."""
        database = schema.__database__
        if self.attr.name in database.values:
            self.valno = database.values[self.attr.name]
        else:
            self.valno = database.add_value(self.attr.name, self.sortable)

    def index(self, schema, name, obj, state):
        value = state.get(self.attr.name)
        prefix = self.prefix or self.attr.name
        if value:
            if self.sortable:
                if self.sortable == 'numeric':
                    value = xapian.sortable_serialise(float(value))
                if self.sortable == 'date':
                    value = value.strftime("%Y%m%d")
                    pre = prefix+'_date'
                    self.check_prefix(schema, prefix=pre)
                    self.add_term(schema, value, prefix=pre)
                if self.sortable == 'datetime':
                    value = value.strftime("%Y%m%d%H%M%S")
                    pre = prefix+'_datetime'
                    self.check_prefix(schema, prefix=pre)
                    self.add_term(schema, value, prefix=pre)
                if self.sortable == 'string':
                    value = str(value)
            else:
                value = dumps(value)
            schema.__document__.add_value(self.valno, value)




import datetime
import xodb
import tempfile
import xapian
from xodb import Schema, Attribute, MultipleValueRangeProcessor

from nose.tools import assert_raises


# These tests and the Attribute description syntax the use is
# deprecated, these will be removed when Attribute style schemas go
# away completely


class Ding(object):
    def __init__(self):
        self.blig = 41
        self.bloog = 42
        self.sez = set([])
        self.canary = []

        self.language = 'en'
        self.name = 'ding'
        self.direction = 'north'
        self.title = 'The young and the database.'
        self.instructions = 'Under the walker'
        self.zig = 3
        self.zag = 6
        self.zork = ['zap', 'zop', 'zern']
        self.zirk = {'jib': 'zoop', 'jab': 'zeep'}
        self.bing = ['boop', 'berp', 'barp']
        self.bang = {'plop': 'This is a borp of a time', 'plip': 'Textual birpism'}
        self.slap = 3
        self.slag = datetime.date(2009, 01, 20)
        self.slup = datetime.datetime(2009, 01, 20, 11, 58, 59)
        self.slat = 'fuzzy'


class DingSchema(Schema):

    def __before__(self, obj):
        obj.canary.append(self)

    name          = Attribute(text=True)
    direction     = Attribute(prefixed=True)
    title         = Attribute(text=True)
    instructions  = Attribute(prefixed=True, text=True)
    zig           = Attribute(value=True)
    zag           = Attribute(value=True, value_parent=zig)
    zork          = Attribute(value=True, sequence=True)
    zirk          = Attribute(value=True, prefixed=True, mapping=True)
    bing          = Attribute(value=True, sequence=True, prefixed=True)
    bang          = Attribute(value=True, prefixed=True, text=True, mapping=True)
    slap          = Attribute(prefixed=True, sortable='numeric')
    slag          = Attribute(prefixed=True, sortable='date')
    slup          = Attribute(prefixed=True, sortable='datetime')
    slat          = Attribute(prefixed=True, sortable='string')
    schlip        = Attribute(required=False, getter=lambda sc, na: "booring")
    canary        = Attribute(volatile=True)
    hipster       = Attribute(default="look")

    def __after__(self, obj, state, doc):
        obj.canary.append(self)


class Dang(object):
    zig = None


class DangSchema(Schema):
    zig = Attribute(prefixed=True, text=True)


class _TestXapianBase(object):

    db_factory = None

    def setup(self):
        self.db = self.db_factory()
        assert self.db.backend.get_doccount() == 0

    def test_no_bogus_db(self):
        assert_raises(TypeError, xodb.Database, None)
        assert_raises(TypeError, xodb.Database, 1)
        assert_raises(TypeError, xodb.Database, True)

    def test_stored_prefixes(self):
        t = tempfile.mkdtemp()
        x = xapian.WritableDatabase(t, xapian.DB_CREATE_OR_OVERWRITE)
        x.set_metadata("_XODB_RP_Foo", "bar")
        x.set_metadata("_XODB_BP_Ding", "dong")
        x.flush()
        xdb = xodb.Database(x)
        assert xdb.relevance_prefixes['Foo'] == 'bar'
        assert xdb.boolean_prefixes['Ding'] == 'dong'

    def test_stored_values(self):
        db = self.db
        assert db.value_count == 0
        assert db.add_value('foo')
        assert db.value_count == 1
        assert db.add_value('foo')
        assert db.value_count == 1
        assert db.add_value('bar')
        assert db.value_count == 2

    def test_values(self):
        d = Ding()
        d.zig = 1
        d.zag = 9
        e = Ding()
        e.zig = 2
        e.zag = 8
        f = Ding()
        f.zig = 3
        e.zag = 7
        self.db.map(Ding, DingSchema)
        for i in [d, e, f]:
            self.db.add(i)

        forwardzig = [d.zig, e.zig, f.zig]
        reversezig = [f.zig, e.zig, d.zig]
        forwardzag = [f.zag, e.zag, d.zag]
        reversezag = [d.zag, e.zag, f.zag]

        results = list(self.db.query('ding', order="zig"))
        assert len(results) == 3
        assert [t.zig for t in results] == forwardzig
        results = list(self.db.query('ding', order="zig", reverse=False))
        assert len(results) == 3
        assert [t.zig for t in results] == forwardzig
        results = list(self.db.query('ding', order="zig", reverse=True))
        assert len(results) == 3
        assert [t.zig for t in results] == reversezig
        results = list(self.db.query('ding', order="zag"))
        assert len(results) == 3
        assert [t.zag for t in results] == forwardzag
        results = list(self.db.query('ding', order="zag", reverse=False))
        assert len(results) == 3
        assert [t.zag for t in results] == forwardzag
        results = list(self.db.query('ding', order="zag", reverse=True))
        assert len(results) == 3
        assert [t.zag for t in results] == reversezag

    def test_schema_text_search(self):
        """ Text searching with unprefixed terms. """
        d = Ding()
        self.db.map(Ding, DingSchema)
        self.db.add(d)
        results = list(self.db.query("young"))
        assert len(results) == 1
        results = list(self.db.query("database"))
        assert len(results) == 1
        results = list(self.db.query("foobar"))
        assert len(results) == 0
        assert self.db.term_freq("young") == 1

    def test_default_value(self):
        d = Ding()
        self.db.map(Ding, DingSchema)
        self.db.add(d)
        results = list(self.db.query("young"))
        assert results[0].hipster == "look"

    def test_required_missing(self):
        class foo(object): pass

        class fooschema(Schema):
            bar = Attribute()

        self.db.map(foo, fooschema)
        d = foo()
        assert_raises(TypeError, self.db.add, d)

    def test_attribute_type(self):
        class foo(object): 
            def __init__(self):
                self.z = 5

        class fooschema(Schema):
            z = Attribute(type=str)

        self.db.map(foo, fooschema)
        d = foo()
        assert_raises(TypeError, self.db.add, d)

    def test_prefixed_text_search(self):
        """ Text searching with prefixed terms. """
        d = Ding()
        self.db.map(Ding, DingSchema)
        self.db.add(d)
        results = list(self.db.query("instructions:walker"))
        assert len(results) == 1
        results = list(self.db.query("instructions:bug"))
        assert len(results) == 0

    def test_prefixed_term_search(self):
        """ Term searching with prefix. """
        d = Ding()
        self.db.map(Ding, DingSchema)
        self.db.add(d)
        results = list(self.db.query("direction:north"))
        assert len(results) == 1
        results = list(self.db.query("direction:south"))
        assert len(results) == 0

    def test_schema_term_search(self):
        """ Term searching without prefix. """
        self.db.map(Ding, DingSchema)
        d = Ding()
        self.db.add(d)
        results = list(self.db.query("ding"))
        assert len(results) == 1
        results = list(self.db.query("dong"))
        assert len(results) == 0

    def test_sequence_terms(self):
        self.db.map(Ding, DingSchema)
        d = Ding()
        self.db.add(d)
        results = list(self.db.query("zap"))
        assert len(results) == 1
        results = list(self.db.query("zop"))
        assert len(results) == 1
        results = list(self.db.query("zern"))
        assert len(results) == 1
        results = list(self.db.query("zowwie"))
        assert len(results) == 0

    def test_sequence_prefixed_terms(self):
        self.db.map(Ding, DingSchema)
        d = Ding()
        self.db.add(d)
        results = list(self.db.query("bing:boop"))
        assert len(results) == 1
        results = list(self.db.query("bing:berp"))
        assert len(results) == 1
        results = list(self.db.query("bing:barp"))
        assert len(results) == 1
        results = list(self.db.query("bing:zowwie"))
        assert len(results) == 0

    def test_mapping_terms(self):
        self.db.map(Ding, DingSchema)
        d = Ding()
        self.db.add(d)
        results = list(self.db.query("zirk_jib:zoop"))
        assert len(results) == 1
        results = list(self.db.query("zirk_jab:zeep"))
        assert len(results) == 1
        results = list(self.db.query("zirk_jib:nono"))
        assert len(results) == 0

    def test_mapping_prefixed_text(self):
        self.db.map(Ding, DingSchema)
        d = Ding()
        self.db.add(d)
        results = list(self.db.query("bang_plop:borp"))
        assert len(results) == 1
        results = list(self.db.query("bang_plip:birpism"))
        assert len(results) == 1
        results = list(self.db.query("bang_plop:nono"))
        assert len(results) == 0

    def test_numeric_sort_value(self):
        self.db.map(Ding, DingSchema)
        d = Ding()
        d.slap = 3
        e = Ding()
        e.slap = 2
        self.db.add(d)
        self.db.add(e)
        results = list(self.db.query("slap:3"))
        assert len(results) == 1
        results = list(self.db.query("slap:3..4"))
        assert len(results) == 1
        results = list(self.db.query("slap:2..4"))
        assert len(results) == 2

    def test_date_and_datetime_sort_value(self):
        self.db.map(Ding, DingSchema)
        d = Ding()
        d.slag = datetime.date(2009, 1, 10)
        d.slup = datetime.datetime(2009, 01, 20, 1, 2, 3)
        e = Ding()
        e.slag = datetime.date(2009, 4, 20)
        e.slup = datetime.datetime(2009, 4, 20, 11, 58, 59)
        self.db.add(d)
        self.db.add(e)
        assert len(list(self.db.query("slag_date:20090420"))) == 1
        assert len(list(self.db.query("slag_date:20090421"))) == 0
        assert len(list(self.db.query("slag_date:200904*"))) == 1
        assert len(list(self.db.query("slag_date:2009*"))) == 2
        results = list(self.db.query("slag:20090201..20090501"))
        assert len(results) == 1
        results = list(self.db.query("slag:20090101..20090501"))
        assert len(results) == 2
        results = list(self.db.query("slag:19700101..20080501"))
        assert len(results) == 0

        assert len(list(self.db.query("slup_datetime:20090420115859"))) == 1
        assert len(list(self.db.query("slup_datetime:20090421115959"))) == 0
        assert len(list(self.db.query("slup_datetime:200904*"))) == 1
        assert len(list(self.db.query("slup_datetime:2009*"))) == 2

        results = list(self.db.query("slup:20090201..20090501"))
        assert len(results) == 1
        results = list(self.db.query("slup:20090101..20090501"))
        assert len(results) == 2
        results = list(self.db.query("slup:19700101..20080501"))
        assert len(results) == 0

    def test_string_sort_value(self):
        self.db.map(Ding, DingSchema)
        d = Ding()
        d.slat = "fuzzybear"
        e = Ding()
        e.slat = "fuzzybunny"
        self.db.add(d)
        self.db.add(e)
        results = list(self.db.query("slat:fuzzybear"))
        assert len(results) == 1
        results = list(self.db.query("slat:fuzzy..fuzzybrine"))
        assert len(results) == 1
        results = list(self.db.query("slat:fuzzy..gump"))
        assert len(results) == 2
        results = list(self.db.query("slat:gump..zed"))
        assert len(results) == 0

    def test_estimate(self):
        self.db.map(Ding, DingSchema)
        d = Ding()
        self.db.add(d)
        assert self.db.estimate("bang_plop:borp") == 1
        assert self.db.backend.get_doccount() == 1

    def test_empty_query(self):
        assert not len(list(self.db.query('')))

    def test_describe_query(self):
        assert self.db.describe_query('foo') == 'Xapian::Query(foo:(pos=1))'

    def test_before_after(self):
        self.db.map(Ding, DingSchema)
        d = Ding()
        self.db.add(d)
        assert len(d.canary) == 2

    def test_before_after_terms(self):
        self.db.map(Ding, DingSchema)
        d = Ding()
        self.db.add(d)
        r = list(self.db.query("booring"))
        assert len(r) == 1

    def test_instance_factory(self):
        self.db = xodb.temp(record_factory=xodb.instance_factory)
        d = Ding()
        self.db.map(Ding, DingSchema)
        i = self.db.add(d)[0]
        e = self.db.get(i)
        assert isinstance(e, Ding)
        assert not hasattr(e, 'blig')
        assert not hasattr(e, 'canary')
        assert e.zig == 3
        assert e.zork == ['zap', 'zop', 'zern']
        assert e.zirk == {'jib': 'zoop', 'jab': 'zeep'}
        assert e.slag == datetime.date(2009, 01, 20)


class TestXapianFile(_TestXapianBase):
    db_factory = staticmethod(xodb.temp)


class TestXapianInMem(_TestXapianBase):
    db_factory = staticmethod(xodb.inmemory)


def test_value_range_processor():
    vp = MultipleValueRangeProcessor(dict(foo=1, bar=2), str.upper)
    assert vp('foo:abc', 'def') == (1, 'ABC', 'DEF')
    assert vp('bar:news', 'def') == (2, 'NEWS', 'DEF')
    assert vp('bar:', 'def') == (2, '', 'DEF')
    assert vp('bar', 'def') == (xapian.BAD_VALUENO, 'bar', 'def')
    assert vp('baz:foo', 'def') == (xapian.BAD_VALUENO, 'baz:foo', 'def')

    qp = xapian.QueryParser()
    db = xodb.temp()
    qp.set_database(db.backend)
    qp.add_valuerangeprocessor(vp)

    query = qp.parse_query('foo:abc..def')
    assert str(query) == 'Xapian::Query(VALUE_RANGE 1 ABC DEF)'

    query = qp.parse_query('bar:abc..def')
    assert str(query) == 'Xapian::Query(VALUE_RANGE 2 ABC DEF)'

    query = qp.parse_query('bar:3..4')
    assert str(query) == 'Xapian::Query(VALUE_RANGE 2 3 4)'

    assert_raises(xapian.QueryParserError, qp.parse_query, 'baz:abc..def')


class TestXapianSuggest(object):

    def setup(self):
        self.db = xodb.temp()
        self.db.map(Dang, DangSchema)
        self.dangs = []
        for value in ('food bar', 'food foo', 'food baz', 'fee fum'):
            d = Dang()
            d.zig = value
            self.db.add(d)
            self.dangs.append(d)

#     def test_suggest(self):
#         result = self.db.suggest('zig:bar')
#         assert result.next()

    def test_similar(self):
        result = self.db.similar(terms=['food', 'spice'])
        dangs = [d.zig for d in result]
        assert len(dangs) == 3
        assert 'fee fum' not in dangs

#     def test_similar_query(self):
#         import pdb; pdb.set_trace()
#         result = self.db.similar('bar')
#         dangs = [d.zig for d in result]
#         assert len(dangs) == 3
#         assert 'fee fum' not in dangs

    def test_similar_query_with_terms(self):
        result = self.db.similar('zig:food', terms=['bar', 'baz'])
        dangs = [d.zig for d in result]
        assert len(dangs) == 2 
        assert 'fee fum' not in dangs
        assert 'food foo' not in dangs

